# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba import cuda, errors, typeof, types
from numba.cuda.testing import unittest, CUDATestCase, skip_on_cudasim
import numpy as np

tested_types = (
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
    np.float16,
    np.float32,
    np.float64,
)

load_operators = (
    (cuda.ldca, "ca"),
    (cuda.ldcg, "cg"),
    (cuda.ldcs, "cs"),
    (cuda.ldlu, "lu"),
    (cuda.ldcv, "cv"),
)
store_operators = (
    (cuda.stcg, "cg"),
    (cuda.stcs, "cs"),
    (cuda.stwb, "wb"),
    (cuda.stwt, "wt"),
)


@skip_on_cudasim("Cache hints not supported on simulator")
class TestCacheHints(CUDATestCase):
    def test_loads(self):
        for operator, modifier in load_operators:

            @cuda.jit
            def f(r, x):
                for i in range(len(r)):
                    r[i] = operator(x, i)

            def f_ptr(ptr, r, n):
                for i in range(n):
                    r[i] = operator(ptr, i)

            for ty in tested_types:
                x = np.arange(5).astype(ty)
                numba_type = typeof(x)
                bitwidth = numba_type.dtype.bitwidth

                with self.subTest(operator=operator, ty=ty, input_type="array"):
                    r = np.zeros_like(x)
                    f[1, 1](r, x)
                    np.testing.assert_equal(r, x)

                    # Check PTX contains a cache-policy load instruction
                    sig = (numba_type, numba_type)
                    ptx = f.inspect_asm(signature=sig)

                    self.assertIn(f"ld.global.{modifier}.b{bitwidth}", ptx)

                with self.subTest(
                    operator=operator, ty=ty, input_type="cpointer"
                ):
                    ptr_type = types.CPointer(numba_type.dtype)

                    sig = (ptr_type, numba_type, types.intp)
                    ptx, _ = cuda.compile_ptx(f_ptr, sig)

                    self.assertIn(f"ld.global.{modifier}.b{bitwidth}", ptx)

    def test_loads_2d(self):
        for operator, modifier in load_operators:

            @cuda.jit
            def f(r, x):
                for i in range(x.shape[0]):
                    for j in range(x.shape[1]):
                        r[i, j] = operator(x, (i, j))

            for ty in tested_types:
                x = np.arange(12).reshape(3, 4).astype(ty)
                numba_type = typeof(x)
                bitwidth = numba_type.dtype.bitwidth

                with self.subTest(operator=operator, ty=ty, dims=2):
                    r = np.zeros_like(x)
                    f[1, 1](r, x)
                    np.testing.assert_equal(r, x)

                    sig = (numba_type, numba_type)
                    ptx = f.inspect_asm(signature=sig)

                    self.assertIn(f"ld.global.{modifier}.b{bitwidth}", ptx)

    def test_stores(self):
        for operator, modifier in store_operators:

            @cuda.jit
            def f(r, x):
                for i in range(len(r)):
                    operator(r, i, x[i])

            def f_ptr(ptr, values, n):
                for i in range(n):
                    operator(ptr, i, values[i])

            for ty in tested_types:
                x = np.arange(5).astype(ty)
                numba_type = typeof(x)
                bitwidth = numba_type.dtype.bitwidth

                with self.subTest(operator=operator, ty=ty):
                    r = np.zeros_like(x)
                    f[1, 1](r, x)
                    np.testing.assert_equal(r, x)

                    # Check PTX contains a cache-policy store instruction
                    sig = (numba_type, numba_type)
                    ptx = f.inspect_asm(signature=sig)

                    self.assertIn(f"st.global.{modifier}.b{bitwidth}", ptx)

                with self.subTest(
                    operator=operator, ty=ty, input_type="cpointer"
                ):
                    ptr_type = types.CPointer(numba_type.dtype)

                    sig = (ptr_type, numba_type, types.intp)
                    ptx, _ = cuda.compile_ptx(f_ptr, sig)

                    self.assertIn(f"st.global.{modifier}.b{bitwidth}", ptx)

    def test_stores_2d(self):
        for operator, modifier in store_operators:

            @cuda.jit
            def f(r, x):
                for i in range(x.shape[0]):
                    for j in range(x.shape[1]):
                        operator(r, (i, j), x[i, j])

            for ty in tested_types:
                x = np.arange(12).reshape(3, 4).astype(ty)
                numba_type = typeof(x)
                bitwidth = numba_type.dtype.bitwidth

                with self.subTest(operator=operator, ty=ty):
                    r = np.zeros_like(x)
                    f[1, 1](r, x)
                    np.testing.assert_equal(r, x)

                    sig = (numba_type, numba_type)
                    ptx = f.inspect_asm(signature=sig)

                    self.assertIn(f"st.global.{modifier}.b{bitwidth}", ptx)

    def test_bad_indices(self):
        def float_indices(x):
            cuda.ldcs(x, 1.0)

        sig_1d = (types.float32[::1],)

        msg = "float64 is not a valid index"
        with self.assertRaisesRegex(errors.TypingError, msg):
            cuda.compile_ptx(float_indices, sig_1d)

        def too_long_indices(x):
            cuda.ldcs(x, (1, 2))

        msg = "Expected 1 indices, got 2"
        with self.assertRaisesRegex(errors.TypingError, msg):
            cuda.compile_ptx(too_long_indices, sig_1d)

        def too_short_indices_scalar(x):
            cuda.ldcs(x, 1)

        def too_short_indices_tuple(x):
            cuda.ldcs(x, (1,))

        sig_2d = (types.float32[:, ::1],)

        msg = "Expected 2 indices, got a scalar"
        with self.assertRaisesRegex(errors.TypingError, msg):
            cuda.compile_ptx(too_short_indices_scalar, sig_2d)

        msg = "Expected 2 indices, got 1"
        with self.assertRaisesRegex(errors.TypingError, msg):
            cuda.compile_ptx(too_short_indices_tuple, sig_2d)

    def test_unsupported_dtypes(self):
        def load_complex(r, x):
            r[0] = cuda.ldcs(x, 0)

        # complex types are not supported (not integer or float)
        numba_type = typeof(np.zeros(1, dtype=np.complex64))
        sig = (numba_type, numba_type)

        msg = "ldcs requires array of integer or float type"
        with self.assertRaisesRegex(errors.TypingError, msg):
            cuda.compile_ptx(load_complex, sig)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import re
import cffi

import numpy as np

from io import StringIO
from numba import cuda
from numba.cuda import float32, float64, int32, intp
from numba.cuda.types import float16, CPointer
from numba.cuda import declare_device
from numba.cuda.testing import unittest, CUDATestCase
from numba.cuda.testing import (
    skip_on_cudasim,
    skip_with_nvdisasm,
    skip_without_nvdisasm,
    skip_if_nvjitlink_missing,
)


@skip_on_cudasim("Simulator does not generate code to be inspected")
class TestInspect(CUDATestCase):
    @property
    def cc(self):
        return cuda.current_context().device.compute_capability

    def test_monotyped(self):
        sig = (float32, int32)

        @cuda.jit(sig)
        def foo(x, y):
            """
            // LLVM: define void
            // LLVM-SAME: foo
            // LLVM-LABEL: entry:
            // LLVM-NEXT:         br label %"[[VAL_0:.*]]"
            // LLVM-NEXT:       [[VAL_0]]:
            // LLVM-NEXT:         ret void

            // ASM: Generated by NVIDIA NVVM Compiler
            // ASM: foo
            """
            pass

        file = StringIO()
        foo.inspect_types(file=file)
        typeanno = file.getvalue()
        # Function name in annotation
        self.assertIn("foo", typeanno)
        # Signature in annotation
        self.assertIn("(float32, int32)", typeanno)
        file.close()

        self.assertFileCheckLLVM(foo, sig)
        self.assertFileCheckAsm(foo, sig)

    def test_polytyped(self):
        @cuda.jit
        def foo(x, y):
            """
            // LLVM: define void
            // LLVM-SAME: foo
            // LLVM_INT-SAME: i64
            // LLVM_INT-SAME: i64
            // LLVM_FLOAT-SAME: double
            // LLVM_FLOAT-SAME: double

            // ASM: Generated by NVIDIA NVVM Compiler
            // ASM: .visible
            // ASM-SAME: .entry
            // ASM-SAME: foo
            """
            pass

        foo[1, 1](1, 1)
        foo[1, 1](1.2, 2.4)

        int_sig = (intp, intp)
        float_sig = (float64, float64)

        self.assertFileCheckLLVM(
            foo, int_sig, check_prefixes=["LLVM", "LLVM_INT"]
        )
        self.assertFileCheckAsm(foo, int_sig, check_prefixes=["ASM"])
        self.assertFileCheckLLVM(
            foo, float_sig, check_prefixes=["LLVM", "LLVM_FLOAT"]
        )
        self.assertFileCheckAsm(foo, float_sig, check_prefixes=["ASM"])

        file = StringIO()
        foo.inspect_types(file=file)
        typeanno = file.getvalue()
        file.close()
        # Signature in annotation
        self.assertIn("({0}, {0})".format(intp), typeanno)
        self.assertIn("(float64, float64)", typeanno)

        # Signature in LLVM dict
        llvmirs = foo.inspect_llvm()
        self.assertEqual(
            2,
            len(llvmirs),
        )
        self.assertIn((intp, intp), llvmirs)
        self.assertIn((float64, float64), llvmirs)

        asmdict = foo.inspect_asm()

        # Signature in assembly dict
        self.assertEqual(
            2,
            len(asmdict),
        )
        self.assertIn((intp, intp), asmdict)
        self.assertIn((float64, float64), asmdict)

    def _test_inspect_sass(self, kernel, name, sass):
        # Ensure function appears in output
        seen_function = False
        for line in sass.split():
            if ".text" in line and name in line:
                seen_function = True
        self.assertTrue(seen_function)

        self.assertRegex(sass, r'//## File ".*/test_inspect.py", line [0-9]')

        # Some instructions common to all supported architectures that should
        # appear in the output
        self.assertIn("S2R", sass)  # Special register to register
        self.assertIn("BRA", sass)  # Branch
        self.assertIn("EXIT", sass)  # Exit program

    @skip_on_cudasim("Simulator does not generate code to be inspected")
    @skip_if_nvjitlink_missing("nvJitLink is required for LTO")
    def test_inspect_lto_asm(self):
        ffi = cffi.FFI()

        ext = cuda.CUSource("""
            #include <cuda_fp16.h>
            extern "C"
            __device__ int add_f2_f2(__half * res, __half * a, __half *b) {
                *res = *a + *b;
                return 0;
            }
            """)

        add = declare_device(
            "add_f2_f2",
            float16(CPointer(float16), CPointer(float16)),
            link=ext,
        )

        @cuda.jit
        def k(arr):
            local_arr = cuda.local.array(shape=1, dtype=np.float16)
            local_arr2 = cuda.local.array(shape=1, dtype=np.float16)
            local_arr[0] = 1
            local_arr2[0] = 2

            ptr = ffi.from_buffer(local_arr)
            ptr2 = ffi.from_buffer(local_arr2)

            arr[0] = add(ptr, ptr2)

        arr = np.array([0], dtype=np.float16)

        k[1, 1](arr)

        allasms = k.inspect_asm()
        asm = next(iter(allasms.values()))

        regex = re.compile(r"call(.|\n)*add_f2_f2")
        self.assertRegex(asm, regex)

        all_ext_asms = k.inspect_lto_ptx()
        lto_asm = next(iter(all_ext_asms.values()))

        self.assertIn("add.f16", lto_asm)
        self.assertNotIn("call", lto_asm)

        np.testing.assert_equal(arr[0], np.float16(1) + np.float16(2))

    def skip_on_cuda_version_issues(self):
        # FIXME: This should be unskipped once the cause of certain nvdisasm
        # versions failing to dump SASS with certain driver / nvJitLink
        # versions is understood
        self.skipTest(
            "Relocation information required for analysis not preserved"
        )

    @skip_without_nvdisasm("nvdisasm needed for inspect_sass()")
    def test_inspect_sass_eager(self):
        self.skip_on_cuda_version_issues()

        sig = (float32[::1], int32[::1])

        @cuda.jit(sig, lineinfo=True)
        def add(x, y):
            i = cuda.grid(1)
            if i < len(x):
                x[i] += y[i]

        self._test_inspect_sass(add, "add", add.inspect_sass(sig))

    @skip_without_nvdisasm("nvdisasm needed for inspect_sass()")
    def test_inspect_sass_lazy(self):
        self.skip_on_cuda_version_issues()

        @cuda.jit(lineinfo=True)
        def add(x, y):
            i = cuda.grid(1)
            if i < len(x):
                x[i] += y[i]

        x = np.arange(10).astype(np.int32)
        y = np.arange(10).astype(np.float32)
        add[1, 10](x, y)

        signature = (int32[::1], float32[::1])
        self._test_inspect_sass(add, "add", add.inspect_sass(signature))

    @skip_with_nvdisasm(
        "Missing nvdisasm exception only generated when it is not present"
    )
    def test_inspect_sass_nvdisasm_missing(self):
        @cuda.jit((float32[::1],))
        def f(x):
            x[0] = 0

        with self.assertRaises(RuntimeError) as raises:
            f.inspect_sass()

        self.assertIn("nvdisasm has not been found", str(raises.exception))

    @skip_without_nvdisasm("nvdisasm needed for inspect_sass_cfg()")
    def test_inspect_sass_cfg(self):
        self.skip_on_cuda_version_issues()

        sig = (float32[::1], int32[::1])

        @cuda.jit(sig)
        def add(x, y):
            i = cuda.grid(1)
            if i < len(x):
                x[i] += y[i]

        self.assertRegex(
            add.inspect_sass_cfg(signature=sig), r"digraph\s*\w\s*{(.|\n)*\n}"
        )


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import numpy as np
from numba import cuda
from numba.cuda.testing import unittest, CUDATestCase
from cuda.core.experimental import Device
from numba.cuda.testing import skip_on_cudasim


class TestCudaEvent(CUDATestCase):
    def test_event_elapsed(self):
        N = 32
        dary = cuda.device_array(N, dtype=np.double)
        evtstart = cuda.event()
        evtend = cuda.event()

        evtstart.record()
        cuda.to_device(np.arange(N, dtype=np.double), to=dary)
        evtend.record()
        evtend.wait()
        evtend.synchronize()
        # Exercise the code path
        evtstart.elapsed_time(evtend)

    def test_event_elapsed_stream(self):
        stream = cuda.stream()
        self.event_elapsed_inner(stream)

    @skip_on_cudasim("Testing cuda.core events requires driver")
    def test_event_elapsed_cuda_core_stream(self):
        dev = Device()
        dev.set_current()
        stream = dev.create_stream()
        self.event_elapsed_inner(stream)

    def event_elapsed_inner(self, stream):
        N = 32
        dary = cuda.device_array(N, dtype=np.double)
        evtstart = cuda.event()
        evtend = cuda.event()

        evtstart.record(stream=stream)
        cuda.to_device(np.arange(N, dtype=np.double), to=dary, stream=stream)
        evtend.record(stream=stream)
        evtend.wait(stream=stream)
        evtend.synchronize()
        # Exercise the code path
        evtstart.elapsed_time(evtend)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import numpy as np

from collections import namedtuple
from functools import partial
from itertools import product
from numba.cuda import vectorize as cuda_vectorize
from numba import cuda, vectorize as numba_vectorize
from numba.cuda.types import int32, float32, float64
from numba.cuda.cudadrv.driver import CudaAPIError, driver
from numba.cuda.testing import skip_on_cudasim
from numba.cuda.testing import CUDATestCase
import unittest


# Signatures to test with - these are all homogeneous in dtype, so the output
# dtype should match the input dtype - the output should not have been cast
# upwards, as reported in #8400: https://github.com/numba/numba/issues/8400
signatures = [
    int32(int32, int32),
    float32(float32, float32),
    float64(float64, float64),
]

# The order here is chosen such that each subsequent dtype might have been
# casted to a previously-used dtype. This is unlikely to be an issue for CUDA,
# but there might be future circumstances in which it becomes relevant, perhaps
# if it supported Dynamic UFuncs, and we want to ensure that an implementation
# for a the given dtype is used rather than casting the input upwards.
dtypes = (np.float64, np.float32, np.int32)

# NumPy ndarray orders
orders = ("C", "F")

# Input sizes corresponding to operations:
# - Less than one warp,
# - Less than one block,
# - Greater than one block (i.e. many blocks)
input_sizes = (8, 100, 2**10 + 1)

# Vectorize functions to test
# cuda.vectorize doesn't need target parameter, numba.vectorize needs target="cuda"
vectorize_funcs = [cuda_vectorize, partial(numba_vectorize, target="cuda")]


@skip_on_cudasim("ufunc API unsupported in the simulator")
class TestCUDAVectorize(CUDATestCase):
    # Presumably chosen as an odd number unlikely to coincide with the total
    # thread count, and large enough to ensure a significant number of blocks
    # are used.
    N = 1000001

    def test_scalar(self):
        for vectorize in vectorize_funcs:

            @vectorize(signatures)
            def vector_add(a, b):
                return a + b

            a = 1.2
            b = 2.3
            c = vector_add(a, b)
            self.assertEqual(c, a + b)

    def test_1d(self):
        for vectorize in vectorize_funcs:

            @vectorize(signatures)
            def vector_add(a, b):
                return a + b

            for ty in dtypes:
                data = np.array(np.random.random(self.N), dtype=ty)
                expected = np.add(data, data)
                actual = vector_add(data, data)
                np.testing.assert_allclose(expected, actual)
                self.assertEqual(actual.dtype, ty)

    def test_1d_async(self):
        for vectorize in vectorize_funcs:

            @vectorize(signatures)
            def vector_add(a, b):
                return a + b

            stream = cuda.stream()

            for ty in dtypes:
                data = np.array(np.random.random(self.N), dtype=ty)
                device_data = cuda.to_device(data, stream)

                dresult = vector_add(device_data, device_data, stream=stream)
                actual = dresult.copy_to_host()

                expected = np.add(data, data)

                np.testing.assert_allclose(expected, actual)
                self.assertEqual(actual.dtype, ty)

    def test_nd(self):
        for vectorize in vectorize_funcs:

            @vectorize(signatures)
            def vector_add(a, b):
                return a + b

            for nd, dtype, order in product(range(1, 8), dtypes, orders):
                shape = (4,) * nd
                data = np.random.random(shape).astype(dtype)
                data2 = np.array(data.T, order=order)

                expected = data + data2
                actual = vector_add(data, data2)
                np.testing.assert_allclose(expected, actual)
                self.assertEqual(actual.dtype, dtype)

    def test_output_arg(self):
        for vectorize in vectorize_funcs:

            @vectorize(signatures)
            def vector_add(a, b):
                return a + b

            A = np.arange(10, dtype=np.float32)
            B = np.arange(10, dtype=np.float32)

            expected = A + B
            actual = np.empty_like(A)
            vector_add(A, B, out=actual)

            np.testing.assert_allclose(expected, actual)
            self.assertEqual(expected.dtype, actual.dtype)

    def test_reduce(self):
        for vectorize in vectorize_funcs:

            @vectorize(signatures)
            def vector_add(a, b):
                return a + b

            dtype = np.int32

            for n in input_sizes:
                x = np.arange(n, dtype=dtype)
                expected = np.add.reduce(x)
                actual = vector_add.reduce(x)
                np.testing.assert_allclose(expected, actual)
                # np.add.reduce is special-cased to return an int64 for any int
                # arguments, so we can't compare against its returned dtype when
                # we're checking the general reduce machinery (which just happens
                # to be using addition). Instead, compare against the input dtype.
                self.assertEqual(dtype, actual.dtype)

    def test_reduce_async(self):
        for vectorize in vectorize_funcs:

            @vectorize(signatures)
            def vector_add(a, b):
                return a + b

            stream = cuda.stream()
            dtype = np.int32

            for n in input_sizes:
                x = np.arange(n, dtype=dtype)
                expected = np.add.reduce(x)
                dx = cuda.to_device(x, stream)
                actual = vector_add.reduce(dx, stream=stream)
                np.testing.assert_allclose(expected, actual)
                # Compare against the input dtype as in test_reduce().
                self.assertEqual(dtype, actual.dtype)

    def test_manual_transfer(self):
        for vectorize in vectorize_funcs:

            @vectorize(signatures)
            def vector_add(a, b):
                return a + b

            n = 10
            x = np.arange(n, dtype=np.int32)
            dx = cuda.to_device(x)
            expected = x + x
            actual = vector_add(x, dx).copy_to_host()
            np.testing.assert_equal(expected, actual)
            self.assertEqual(expected.dtype, actual.dtype)

    def test_ufunc_output_2d(self):
        for vectorize in vectorize_funcs:

            @vectorize(signatures)
            def vector_add(a, b):
                return a + b

            n = 10
            x = np.arange(n, dtype=np.int32).reshape(2, 5)
            dx = cuda.to_device(x)
            vector_add(dx, dx, out=dx)

            expected = x + x
            actual = dx.copy_to_host()
            np.testing.assert_equal(expected, actual)
            self.assertEqual(expected.dtype, actual.dtype)

    def check_tuple_arg(self, a, b):
        for vectorize in vectorize_funcs:

            @vectorize(signatures)
            def vector_add(a, b):
                return a + b

            r = vector_add(a, b)
            np.testing.assert_equal(np.asarray(a) + np.asarray(b), r)

    def test_tuple_arg(self):
        a = (1.0, 2.0, 3.0)
        b = (4.0, 5.0, 6.0)
        self.check_tuple_arg(a, b)

    def test_namedtuple_arg(self):
        Point = namedtuple("Point", ("x", "y", "z"))
        a = Point(x=1.0, y=2.0, z=3.0)
        b = Point(x=4.0, y=5.0, z=6.0)
        self.check_tuple_arg(a, b)

    def test_tuple_of_array_arg(self):
        arr = np.arange(10, dtype=np.int32)
        a = (arr, arr + 1)
        b = (arr + 2, arr + 2)
        self.check_tuple_arg(a, b)

    def test_tuple_of_namedtuple_arg(self):
        Point = namedtuple("Point", ("x", "y", "z"))
        a = (Point(x=1.0, y=2.0, z=3.0), Point(x=1.5, y=2.5, z=3.5))
        b = (Point(x=4.0, y=5.0, z=6.0), Point(x=4.5, y=5.5, z=6.5))
        self.check_tuple_arg(a, b)

    def test_namedtuple_of_array_arg(self):
        xs1 = np.arange(10, dtype=np.int32)
        ys1 = xs1 + 2
        xs2 = np.arange(10, dtype=np.int32) * 2
        ys2 = xs2 + 1
        Points = namedtuple("Points", ("xs", "ys"))
        a = Points(xs=xs1, ys=ys1)
        b = Points(xs=xs2, ys=ys2)
        self.check_tuple_arg(a, b)

    def test_name_attribute(self):
        for vectorize in vectorize_funcs:

            @vectorize("f8(f8)")
            def bar(x):
                return x**2

            self.assertEqual(bar.__name__, "bar")

    def test_no_transfer_for_device_data(self):
        for vectorize in vectorize_funcs:
            # Initialize test data on the device prior to banning host <-> device
            # transfer

            noise = np.random.randn(1, 3, 64, 64).astype(np.float32)
            noise = cuda.to_device(noise)

            # A mock of a CUDA function that always raises a CudaAPIError

            def raising_transfer(*args, **kwargs):
                raise CudaAPIError(999, "Transfer not allowed")

            # Use the mock for transfers between the host and device

            old_HtoD = getattr(driver, "cuMemcpyHtoD", None)
            old_DtoH = getattr(driver, "cuMemcpyDtoH", None)

            setattr(driver, "cuMemcpyHtoD", raising_transfer)
            setattr(driver, "cuMemcpyDtoH", raising_transfer)

            # Ensure that the mock functions are working as expected

            with self.assertRaisesRegex(CudaAPIError, "Transfer not allowed"):
                noise.copy_to_host()

            with self.assertRaisesRegex(CudaAPIError, "Transfer not allowed"):
                cuda.to_device([1])

            try:
                # Check that defining and calling a ufunc with data on the device
                # induces no transfers

                @vectorize(["float32(float32)"])
                def func(noise):
                    return noise + 1.0

                func(noise)
            finally:
                # Replace our mocks with the original implementations. If there was
                # no original implementation, simply remove ours.

                if old_HtoD is not None:
                    setattr(driver, "cuMemcpyHtoD", old_HtoD)
                else:
                    del driver.cuMemcpyHtoD
                if old_DtoH is not None:
                    setattr(driver, "cuMemcpyDtoH", old_DtoH)
                else:
                    del driver.cuMemcpyDtoH


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
"""
Tests for SSA reconstruction
"""

import sys
import copy
import logging

import numpy as np

from numba.cuda import types
from numba import cuda
from numba.cuda import jit
from numba.cuda.core import errors

from numba.cuda.extending import overload
from numba.cuda.tests.support import override_config
from numba.cuda.testing import CUDATestCase, skip_on_cudasim


_DEBUG = False

if _DEBUG:
    # Enable debug logger on SSA reconstruction
    ssa_logger = logging.getLogger("numba.cuda.core.ssa")
    ssa_logger.setLevel(level=logging.DEBUG)
    ssa_logger.addHandler(logging.StreamHandler(sys.stderr))


class SSABaseTest(CUDATestCase):
    """
    This class comes from numba tests, but has been modified to work with CUDA kernels.
    Return values were replaced by output arrays, and tuple returns assign to elements of the output array.
    """

    def check_func(self, func, result_array, *args):
        # For CUDA kernels, we need to create output arrays and call with [1,1] launch config
        # Create GPU array with same shape as expected result array
        gpu_result_array = cuda.to_device(np.zeros_like(result_array))

        # Call the CUDA kernel
        func[1, 1](gpu_result_array, *copy.deepcopy(args))
        gpu_result = gpu_result_array.copy_to_host()

        # Call the original Python function for expected result
        cpu_result = np.zeros_like(result_array)
        func.py_func(cpu_result, *copy.deepcopy(args))

        # Compare all results
        np.testing.assert_array_equal(gpu_result, cpu_result)


class TestSSA(SSABaseTest):
    """
    Contains tests to help isolate problems in SSA
    """

    def test_argument_name_reused(self):
        @jit
        def foo(result, x):
            x += 1
            result[0] = x

        self.check_func(foo, np.array([124.0]), 123)

    def test_if_else_redefine(self):
        @jit
        def foo(result, x, y):
            z = x * y
            if x < y:
                z = x
            else:
                z = y
            result[0] = z

        self.check_func(foo, np.array([2.0]), 3, 2)
        self.check_func(foo, np.array([2.0]), 2, 3)

    def test_sum_loop(self):
        @jit
        def foo(result, n):
            c = 0
            for i in range(n):
                c += i
            result[0] = c

        self.check_func(foo, np.array([0.0]), 0)
        self.check_func(foo, np.array([45.0]), 10)

    def test_sum_loop_2vars(self):
        @jit
        def foo(result, n):
            c = 0
            d = n
            for i in range(n):
                c += i
                d += n
            result[0] = c
            result[1] = d

        self.check_func(foo, np.array([0.0, 0.0]), 0)
        self.check_func(foo, np.array([45.0, 110.0]), 10)

    def test_sum_2d_loop(self):
        @jit
        def foo(result, n):
            c = 0
            for i in range(n):
                for j in range(n):
                    c += j
                c += i
            result[0] = c

        self.check_func(foo, np.array([0.0]), 0)
        self.check_func(foo, np.array([495.0]), 10)

    def check_undefined_var(self, should_warn):
        @jit
        def foo(result, n):
            if n:
                if n > 0:
                    c = 0
                result[0] = c
            else:
                # variable c is not defined in this branch
                c += 1
                result[0] = c

        if should_warn:
            with self.assertWarns(errors.NumbaWarning) as warns:
                # n=1 so we won't actually run the branch with the uninitialized
                self.check_func(foo, np.array([0]), 1)
            self.assertIn(
                "Detected uninitialized variable c", str(warns.warning)
            )
        else:
            self.check_func(foo, np.array([0]), 1)

        with self.assertRaises(UnboundLocalError):
            result = np.array([0])
            foo.py_func(result, 0)

    @skip_on_cudasim(
        "Numba variable warnings are not supported in the simulator"
    )
    def test_undefined_var(self):
        with override_config("ALWAYS_WARN_UNINIT_VAR", 0):
            self.check_undefined_var(should_warn=False)
        with override_config("ALWAYS_WARN_UNINIT_VAR", 1):
            self.check_undefined_var(should_warn=True)

    def test_phi_propagation(self):
        @jit
        def foo(result, actions):
            n = 1

            i = 0
            ct = 0
            while n > 0 and i < len(actions):
                n -= 1

                while actions[i]:
                    if actions[i]:
                        if actions[i]:
                            n += 10
                        actions[i] -= 1
                    else:
                        if actions[i]:
                            n += 20
                        actions[i] += 1

                    ct += n
                ct += n
            result[0] = ct
            result[1] = n

        self.check_func(foo, np.array([1, 2]), np.array([1, 2]))

    def test_unhandled_undefined(self):
        @cuda.jit
        def function1(arg1, arg2, arg3, arg4, arg5):
            # This function is auto-generated.
            if arg1:
                var1 = arg2
                var2 = arg3
                var3 = var2
                var4 = arg1
                return
            else:
                if arg2:
                    if arg4:
                        var5 = arg4  # noqa: F841
                        return
                    else:
                        var6 = var4
                        return
                    return var6
                else:
                    if arg5:
                        if var1:
                            if arg5:
                                var1 = var6
                                return
                            else:
                                var7 = arg2  # noqa: F841
                                return arg2
                            return
                        else:
                            if var2:
                                arg5 = arg2
                                return arg1
                            else:
                                var6 = var3
                                return var4
                            return
                        return
                    else:
                        var8 = var1
                        return
                    return var8
                var9 = var3  # noqa: F841
                var10 = arg5  # noqa: F841
                return var1

        NONE_SENTINEL = 99

        @cuda.jit
        def function1_caller(result, arg1, arg2, arg3, arg4, arg5):
            retval = function1(arg1, arg2, arg3, arg4, arg5)
            if retval is None:
                result[0] = NONE_SENTINEL
            else:
                result[0] = retval

        # The argument values is not critical for re-creating the bug
        # because the bug is in compile-time.

        expect = function1.py_func(2, 3, 6, 0, 7)
        if expect is None:
            expect = NONE_SENTINEL
        result = np.zeros(1, dtype=np.int64)
        function1_caller[1, 1](result, 2, 3, 6, 0, 7)
        got = result[0]
        self.assertEqual(expect, got)


class TestReportedSSAIssues(SSABaseTest):
    # Tests from issues
    # https://github.com/numba/numba/issues?q=is%3Aopen+is%3Aissue+label%3ASSA

    def test_issue2194(self):
        @jit
        def foo(result, V):
            s = np.uint32(1)

            for i in range(s):
                V[i] = 1
            for i in range(s, 1):
                pass
            result[0] = V[0]

        V = np.empty(1)
        self.check_func(foo, np.array([1.0]), V)

    def test_issue3094(self):
        @jit
        def foo(result, pred):
            if pred:
                x = 1
            else:
                x = 0
            result[0] = x

        self.check_func(foo, np.array([0]), False)

    def test_issue3931(self):
        @jit
        def foo(result, arr):
            for i in range(1):
                arr = arr.reshape(3 * 2)
                arr = arr.reshape(3, 2)
            # Copy result array elements
            for i in range(arr.shape[0]):
                for j in range(arr.shape[1]):
                    result[i, j] = arr[i, j]

        result_gpu = np.zeros((3, 2))
        self.check_func(foo, result_gpu, np.zeros((3, 2)))

    def test_issue3976(self):
        def overload_this(a):
            return 42

        @jit
        def foo(result, a):
            if a:
                s = 5
                s = overload_this(s)
            else:
                s = 99

            result[0] = s

        @overload(overload_this)
        def ol(a):
            return overload_this

        self.check_func(foo, np.array([42]), True)

    def test_issue3979(self):
        @jit
        def foo(result, A, B):
            x = A[0]
            y = B[0]
            for i in A:
                x = i
            for i in B:
                y = i
            result[0] = x
            result[1] = y

        self.check_func(
            foo, np.array([2, 4]), np.array([1, 2]), np.array([3, 4])
        )

    def test_issue5219(self):
        def overload_this(a, b=None):
            if isinstance(b, tuple):
                b = b[0]
            return b

        @overload(overload_this)
        def ol(a, b=None):
            b_is_tuple = isinstance(b, (types.Tuple, types.UniTuple))

            def impl(a, b=None):
                if b_is_tuple is True:
                    b = b[0]
                return b

            return impl

        @jit
        def test_tuple(result, a, b):
            result[0] = overload_this(a, b)

        self.check_func(test_tuple, np.array([2]), 1, (2,))

    def test_issue5223(self):
        @jit
        def bar(result, x):
            if len(x) == 5:
                for i in range(len(x)):
                    result[i] = x[i]
            else:
                # Manual copy since .copy() not available in CUDA
                for i in range(len(x)):
                    result[i] = x[i] + 1

        a = np.ones(5)
        a.flags.writeable = False
        expected = np.ones(5)  # Since len(a) == 5, it should return unchanged
        self.check_func(bar, expected, a)

    def test_issue5243(self):
        @jit
        def foo(result, q, lin):
            stencil_val = 0.0  # noqa: F841
            stencil_val = q[0, 0]  # noqa: F841
            result[0] = lin[0]

        lin = np.array([0.1, 0.6, 0.3])
        self.check_func(foo, np.array([0.1]), np.zeros((2, 2)), lin)

    def test_issue5482_missing_variable_init(self):
        # Test error that lowering fails because variable is missing
        # a definition before use.
        @jit
        def foo(result, x, v, n):
            for i in range(n):
                if i == 0:
                    if i == x:
                        pass
                    else:
                        problematic = v
                else:
                    if i == x:
                        pass
                    else:
                        problematic = problematic + v
            result[0] = problematic

        self.check_func(foo, np.array([10]), 1, 5, 3)

    def test_issue5493_unneeded_phi(self):
        # Test error that unneeded phi is inserted because variable does not
        # have a dominance definition.
        data = (np.ones(2), np.ones(2))
        A = np.ones(1)
        B = np.ones(1)

        @jit
        def foo(res, m, n, data):
            if len(data) == 1:
                v0 = data[0]
            else:
                v0 = data[0]
                # Unneeded PHI node for `problematic` would be placed here
                for _ in range(1, len(data)):
                    v0[0] += A[0]

            for t in range(1, m):
                for idx in range(n):
                    t = B

                    if idx == 0:
                        if idx == n - 1:
                            pass
                        else:
                            res[0] = t[0]
                    else:
                        if idx == n - 1:
                            pass
                        else:
                            res[0] += t[0]

        self.check_func(foo, np.array([10]), 10, 10, data)

    def test_issue5623_equal_statements_in_same_bb(self):
        def foo(pred, stack):
            i = 0
            c = 1

            if pred is True:
                stack[i] = c
                i += 1
                stack[i] = c
                i += 1

        python = np.array([0, 666])
        foo(True, python)

        nb = np.array([0, 666])

        # Convert to CUDA kernel
        foo_cuda = jit(foo)
        foo_cuda[1, 1](True, nb)

        expect = np.array([1, 1])

        np.testing.assert_array_equal(python, expect)
        np.testing.assert_array_equal(nb, expect)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import numpy as np

from numba import cuda
from numba.cuda import HAS_NUMBA
from numba.cuda.testing import unittest, CUDATestCase, skip_on_cudasim

if HAS_NUMBA:
    from numba.extending import overload


@skip_on_cudasim("Simulator does not support the extension API")
@unittest.skipUnless(HAS_NUMBA, "Tests interoperability with Numba")
class TestNumbaInterop(CUDATestCase):
    def test_overload_inline_always(self):
        # From Issue #624
        def get_42():
            raise NotImplementedError()

        @overload(get_42, target="cuda", inline="always")
        def ol_blas_get_accumulator():
            def impl():
                return 42

            return impl

        @cuda.jit
        def kernel(a):
            a[0] = get_42()

        a = np.empty(1, dtype=np.float32)
        kernel[1, 1](a)
        np.testing.assert_equal(a[0], 42)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

# Expected to run this module as __main__


# Cloudpickle will think this is a dynamic class when this module is __main__
class Klass:
    classvar = None



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import numpy as np
import math

from numba import cuda
from numba.cuda import vectorize, int32, uint32, float32, float64
from numba.cuda.testing import skip_on_cudasim, CUDATestCase
from numba.cuda.tests.support import CheckWarningsMixin

import unittest


pi = math.pi


def sinc(x):
    if x == 0.0:
        return 1.0
    else:
        return math.sin(x * pi) / (pi * x)


def scaled_sinc(x, scale):
    if x == 0.0:
        return scale
    else:
        return scale * (math.sin(x * pi) / (pi * x))


def vector_add(a, b):
    return a + b


class BaseVectorizeDecor(object):
    target = None
    wrapper = None
    funcs = {
        "func1": sinc,
        "func2": scaled_sinc,
        "func3": vector_add,
    }

    @classmethod
    def _run_and_compare(cls, func, sig, A, *args, **kwargs):
        if cls.wrapper is not None:
            func = cls.wrapper(func)
        numba_func = vectorize(sig, target=cls.target)(func)
        numpy_func = np.vectorize(func)
        result = numba_func(A, *args)
        gold = numpy_func(A, *args)
        np.testing.assert_allclose(result, gold, **kwargs)

    def test_1(self):
        sig = ["float64(float64)", "float32(float32)"]
        func = self.funcs["func1"]
        A = np.arange(100, dtype=np.float64)
        self._run_and_compare(func, sig, A)

    def test_2(self):
        sig = [float64(float64), float32(float32)]
        func = self.funcs["func1"]
        A = np.arange(100, dtype=np.float64)
        self._run_and_compare(func, sig, A)

    def test_3(self):
        sig = ["float64(float64, uint32)"]
        func = self.funcs["func2"]
        A = np.arange(100, dtype=np.float64)
        scale = np.uint32(3)
        self._run_and_compare(func, sig, A, scale, atol=1e-8)

    def test_4(self):
        sig = [
            int32(int32, int32),
            uint32(uint32, uint32),
            float32(float32, float32),
            float64(float64, float64),
        ]
        func = self.funcs["func3"]
        A = np.arange(100, dtype=np.float64)
        self._run_and_compare(func, sig, A, A)
        A = A.astype(np.float32)
        self._run_and_compare(func, sig, A, A)
        A = A.astype(np.int32)
        self._run_and_compare(func, sig, A, A)
        A = A.astype(np.uint32)
        self._run_and_compare(func, sig, A, A)


class BaseVectorizeNopythonArg(unittest.TestCase, CheckWarningsMixin):
    """
    Test passing the nopython argument to the vectorize decorator.
    """

    def _test_target_nopython(self, target, warnings, with_sig=True):
        a = np.array([2.0], dtype=np.float32)
        b = np.array([3.0], dtype=np.float32)
        sig = [float32(float32, float32)]
        args = with_sig and [sig] or []
        with self.check_warnings(warnings):
            f = vectorize(*args, target=target, nopython=True)(vector_add)
            f(a, b)


class BaseVectorizeUnrecognizedArg(unittest.TestCase, CheckWarningsMixin):
    """
    Test passing an unrecognized argument to the vectorize decorator.
    """

    def _test_target_unrecognized_arg(self, target, with_sig=True):
        a = np.array([2.0], dtype=np.float32)
        b = np.array([3.0], dtype=np.float32)
        sig = [float32(float32, float32)]
        args = with_sig and [sig] or []
        with self.assertRaises(KeyError) as raises:
            f = vectorize(*args, target=target, nonexistent=2)(vector_add)
            f(a, b)
        self.assertIn("Unrecognized options", str(raises.exception))


@skip_on_cudasim("ufunc API unsupported in the simulator")
class TestVectorizeDecor(CUDATestCase, BaseVectorizeDecor):
    """
    Runs the tests from BaseVectorizeDecor with the CUDA target.
    """

    target = "cuda"


@skip_on_cudasim("ufunc API unsupported in the simulator")
class TestGPUVectorizeBroadcast(CUDATestCase):
    def test_broadcast(self):
        a = np.random.randn(100, 3, 1)
        b = a.transpose(2, 1, 0)

        def fn(a, b):
            return a - b

        @vectorize(["float64(float64,float64)"], target="cuda")
        def fngpu(a, b):
            return a - b

        expect = fn(a, b)
        got = fngpu(a, b)
        np.testing.assert_almost_equal(expect, got)

    def test_device_broadcast(self):
        """
        Same test as .test_broadcast() but with device array as inputs
        """

        a = np.random.randn(100, 3, 1)
        b = a.transpose(2, 1, 0)

        def fn(a, b):
            return a - b

        @vectorize(["float64(float64,float64)"], target="cuda")
        def fngpu(a, b):
            return a - b

        expect = fn(a, b)
        got = fngpu(cuda.to_device(a), cuda.to_device(b))
        np.testing.assert_almost_equal(expect, got.copy_to_host())


@skip_on_cudasim("ufunc API unsupported in the simulator")
class TestVectorizeNopythonArg(BaseVectorizeNopythonArg, CUDATestCase):
    def test_target_cuda_nopython(self):
        warnings = ["nopython kwarg for cuda target is redundant"]
        self._test_target_nopython("cuda", warnings)


@skip_on_cudasim("ufunc API unsupported in the simulator")
class TestVectorizeUnrecognizedArg(BaseVectorizeUnrecognizedArg, CUDATestCase):
    def test_target_cuda_unrecognized_arg(self):
        self._test_target_unrecognized_arg("cuda")


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from io import StringIO
import logging

import unittest
from numba.cuda.core import tracing

logger = logging.getLogger("trace")
logger.setLevel(logging.INFO)

# Make sure tracing is enabled
orig_trace = tracing.trace
tracing.trace = tracing.dotrace


class CapturedTrace:
    """Capture the trace temporarily for validation."""

    def __init__(self):
        self.buffer = StringIO()
        self.handler = logging.StreamHandler(self.buffer)

    def __enter__(self):
        self._handlers = logger.handlers
        self.buffer = StringIO()
        logger.handlers = [logging.StreamHandler(self.buffer)]

    def __exit__(self, type, value, traceback):
        logger.handlers = self._handlers

    def getvalue(self):
        # Depending on how the tests are run, object names may be
        # qualified by their containing module.
        # Remove that to make the trace output independent from the testing mode.
        log = self.buffer.getvalue()
        log = log.replace(__name__ + ".", "")
        return log


class Class(object):
    @tracing.trace
    @classmethod
    def class_method(cls):
        pass

    @tracing.trace
    @staticmethod
    def static_method():
        pass

    __test = None

    def _test_get(self):
        return self.__test

    def _test_set(self, value):
        self.__test = value

    test = tracing.trace(property(_test_get, _test_set))

    @tracing.trace
    def method(self, some, other="value", *args, **kwds):
        pass

    def __repr__(self):
        """Generate a deterministic string for testing."""
        return "<Class instance>"


class Class2(object):
    @classmethod
    def class_method(cls):
        pass

    @staticmethod
    def static_method():
        pass

    __test = None

    @property
    def test(self):
        return self.__test

    @test.setter
    def test(self, value):
        self.__test = value

    def method(self):
        pass

    def __str__(self):
        return "Test(" + str(self.test) + ")"

    def __repr__(self):
        """Generate a deterministic string for testing."""
        return "<Class2 instance>"


@tracing.trace
def test_traced_function():
    # Test the tracing functionality with fixed values
    x, y = 5, 5
    z = True

    a = x + y
    b = x * y
    if z:
        result = a
    else:
        result = b

    # The function should return 10 (5 + 5) when z is True
    assert result == 10


class TestTracing(unittest.TestCase):
    def __init__(self, *args):
        super(TestTracing, self).__init__(*args)

    def setUp(self):
        self.capture = CapturedTrace()

    def tearDown(self):
        del self.capture

    def test_method(self):
        with self.capture:
            Class().method("foo", bar="baz")
        self.assertEqual(
            self.capture.getvalue(),
            ">> Class.method(self=<Class instance>, some='foo', other='value', bar='baz')\n"
            + "<< Class.method\n",
        )

    def test_class_method(self):
        with self.capture:
            Class.class_method()
        self.assertEqual(
            self.capture.getvalue(),
            ">> Class.class_method(cls=<class 'Class'>)\n"
            + "<< Class.class_method\n",
        )

    def test_static_method(self):
        with self.capture:
            Class.static_method()
        self.assertEqual(
            self.capture.getvalue(),
            ">> static_method()\n" + "<< static_method\n",
        )

    def test_property(self):
        with self.capture:
            test = Class()
            test.test = 1
            assert 1 == test.test
        self.assertEqual(
            self.capture.getvalue(),
            ">> Class._test_set(self=<Class instance>, value=1)\n"
            + "<< Class._test_set\n"
            + ">> Class._test_get(self=<Class instance>)\n"
            + "<< Class._test_get -> 1\n",
        )

    def test_function(self):
        with self.capture:
            test_traced_function()
        # The test function should be traced when called
        trace_output = self.capture.getvalue()
        self.assertIn(">> test_traced_function()", trace_output)
        self.assertIn("<< test_traced_function", trace_output)

    @unittest.skip("recursive decoration not yet implemented")
    def test_injected(self):
        with self.capture:
            tracing.trace(Class2, recursive=True)
            Class2.class_method()
            Class2.static_method()
            test = Class2()
            test.test = 1
            assert 1 == test.test
            test.method()

            self.assertEqual(
                self.capture.getvalue(),
                ">> Class2.class_method(cls=<type 'Class2'>)\n"
                + "<< Class2.class_method\n"
                ">> static_method()\n"
                "<< static_method\n",
            )


# Reset tracing to its original value
tracing.trace = orig_trace

if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Test extending types via the numba.extending.* API.
"""

from numba.cuda import jit
from numba.cuda import types
from numba.cuda import HAS_NUMBA

if HAS_NUMBA:
    from numba.core.errors import TypingError, NumbaTypeError
else:
    from numba.cuda.core.errors import TypingError, NumbaTypeError
from numba.cuda.extending import make_attribute_wrapper
from numba.cuda.extending import overload
from numba.cuda.core.imputils import Registry
from numba.cuda.testing import skip_on_cudasim

import unittest

registry = Registry()
lower = registry.lower


def gen_mock_float():
    # Stub to overload, pretending to be `float`. The real `float` function is
    # not used as multiple registrations can collide.
    def mock_float(x):
        pass

    return mock_float


class TestExtTypDummy(unittest.TestCase):
    def setUp(self):
        class DummyType(types.Type):
            def __init__(self):
                super(DummyType, self).__init__(name="Dummy")

        make_attribute_wrapper(DummyType, "value", "value")

        # Store attributes
        self.DummyType = DummyType

    def _add_float_overload(self, mock_float_inst):
        @overload(mock_float_inst)
        def dummy_to_float(x):
            if isinstance(x, self.DummyType):

                def codegen(x):
                    return float(x.value)

                return codegen
            else:
                raise NumbaTypeError("cannot type float({})".format(x))

    @skip_on_cudasim("Simulator does not support extending types")
    def test_overload_float_error_msg(self):
        mock_float = gen_mock_float()
        self._add_float_overload(mock_float)

        @jit
        def foo(x):
            mock_float(x)

        with self.assertRaises(TypingError) as raises:
            foo[1, 1](1j)

        self.assertIn("cannot type float(complex128)", str(raises.exception))



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import unittest

from numba.cuda.testing import CUDATestCase, skip_on_cudasim
from numba.cuda.tests.support import captured_stdout


@skip_on_cudasim("cudasim doesn't support cuda import at non-top-level")
class TestGlobals(CUDATestCase):
    """
    Tests demonstrating how global variables are captured in CUDA kernels.
    """

    def setUp(self):
        # Prevent output from this test showing
        # up when running the test suite
        self._captured_stdout = captured_stdout()
        self._captured_stdout.__enter__()
        super().setUp()

    def tearDown(self):
        # No exception type, value, or traceback
        self._captured_stdout.__exit__(None, None, None)
        super().tearDown()

    def test_ex_globals_constant_capture(self):
        """
        Test demonstrating how global variables are captured as constants.
        """
        # magictoken.ex_globals_constant_capture.begin
        import numpy as np
        from numba import cuda

        TAX_RATE = 0.08
        PRICES = np.array([10.0, 25.0, 5.0, 15.0, 30.0], dtype=np.float64)

        @cuda.jit
        def compute_totals(quantities, totals):
            i = cuda.grid(1)
            if i < totals.size:
                totals[i] = quantities[i] * PRICES[i] * (1 + TAX_RATE)

        d_quantities = cuda.to_device(
            np.array([1, 2, 3, 4, 5], dtype=np.float64)
        )
        d_totals = cuda.device_array(5, dtype=np.float64)

        # First kernel call - compiles and captures values
        compute_totals[1, 32](d_quantities, d_totals)
        print("Value of d_totals:", d_totals.copy_to_host())

        # These modifications have no effect on subsequent kernel calls
        TAX_RATE = 0.10  # noqa: F841
        PRICES[:] = [20.0, 50.0, 10.0, 30.0, 60.0]

        # Second kernel call still uses the original values
        compute_totals[1, 32](d_quantities, d_totals)
        print("Value of d_totals:", d_totals.copy_to_host())
        # magictoken.ex_globals_constant_capture.end

        # Verify the values are the same (original values were captured)
        expected = np.array([10.8, 54.0, 16.2, 64.8, 162.0])
        np.testing.assert_allclose(d_totals.copy_to_host(), expected)

    def test_ex_globals_device_array_capture(self):
        """
        Test demonstrating how global device arrays are captured by pointer.
        """
        # magictoken.ex_globals_device_array_capture.begin
        import numpy as np
        from numba import cuda

        # Global device array - pointer is captured, not data
        PRICES = cuda.to_device(
            np.array([10.0, 25.0, 5.0, 15.0, 30.0], dtype=np.float32)
        )

        @cuda.jit
        def compute_totals(quantities, totals):
            i = cuda.grid(1)
            if i < totals.size:
                totals[i] = quantities[i] * PRICES[i]

        d_quantities = cuda.to_device(
            np.array([1.0, 1.0, 1.0, 1.0, 1.0], dtype=np.float32)
        )
        d_totals = cuda.device_array(5, dtype=np.float32)

        # First kernel call
        compute_totals[1, 32](d_quantities, d_totals)
        print(d_totals.copy_to_host())  # [10. 25.  5. 15. 30.]

        # Mutate the device array in-place
        PRICES.copy_to_device(
            np.array([20.0, 50.0, 10.0, 30.0, 60.0], dtype=np.float32)
        )

        # Second kernel call sees the updated values
        compute_totals[1, 32](d_quantities, d_totals)
        print(d_totals.copy_to_host())  # [20. 50. 10. 30. 60.]
        # magictoken.ex_globals_device_array_capture.end

        # Verify the second call sees updated values
        expected = np.array([20.0, 50.0, 10.0, 30.0, 60.0], dtype=np.float32)
        np.testing.assert_allclose(d_totals.copy_to_host(), expected)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import multiprocessing
import os
import shutil
import unittest
import warnings
import sys
import stat
import subprocess

from numba import cuda
from numba.cuda.core.errors import NumbaWarning
from numba.cuda.testing import (
    CUDATestCase,
    skip_on_cudasim,
    skip_unless_cc_60,
    skip_if_cudadevrt_missing,
    test_data_dir,
    skip_on_standalone_numba_cuda,
)
from numba.cuda.tests.support import (
    TestCase,
    temp_directory,
    import_dynamic,
)
import numpy as np
from pickle import PicklingError

# Module-level global for testing that caching rejects global device arrays
GLOBAL_DEVICE_ARRAY = None


class BaseCacheTest(TestCase):
    # The source file that will be copied
    usecases_file = None
    # Make sure this doesn't conflict with another module
    modname = None

    def setUp(self):
        self.tempdir = temp_directory("test_cache")
        sys.path.insert(0, self.tempdir)
        self.modfile = os.path.join(self.tempdir, self.modname + ".py")
        self.cache_dir = os.path.join(self.tempdir, "__pycache__")
        shutil.copy(self.usecases_file, self.modfile)
        os.chmod(self.modfile, stat.S_IREAD | stat.S_IWRITE)
        self.maxDiff = None

    def tearDown(self):
        sys.modules.pop(self.modname, None)
        sys.path.remove(self.tempdir)

    def import_module(self):
        # Import a fresh version of the test module.  All jitted functions
        # in the test module will start anew and load overloads from
        # the on-disk cache if possible.
        old = sys.modules.pop(self.modname, None)
        if old is not None:
            # Make sure cached bytecode is removed
            cached = [old.__cached__]
            for fn in cached:
                try:
                    os.unlink(fn)
                except FileNotFoundError:
                    pass
        mod = import_dynamic(self.modname)
        self.assertEqual(mod.__file__.rstrip("co"), self.modfile)
        return mod

    def cache_contents(self):
        try:
            return [
                fn
                for fn in os.listdir(self.cache_dir)
                if not fn.endswith((".pyc", ".pyo"))
            ]
        except FileNotFoundError:
            return []

    def get_cache_mtimes(self):
        return dict(
            (fn, os.path.getmtime(os.path.join(self.cache_dir, fn)))
            for fn in sorted(self.cache_contents())
        )

    def check_pycache(self, n):
        c = self.cache_contents()
        self.assertEqual(len(c), n, c)

    def dummy_test(self):
        pass


class DispatcherCacheUsecasesTest(BaseCacheTest):
    here = os.path.dirname(__file__)
    usecases_file = os.path.join(here, "cache_usecases.py")
    modname = "dispatcher_caching_test_fodder"

    def run_in_separate_process(self, *, envvars={}):
        # Cached functions can be run from a distinct process.
        # Also stresses issue #1603: uncached function calling cached function
        # shouldn't fail compiling.
        code = """if 1:
            import sys

            sys.path.insert(0, %(tempdir)r)
            mod = __import__(%(modname)r)
            mod.self_test()
            """ % dict(tempdir=self.tempdir, modname=self.modname)

        subp_env = os.environ.copy()
        subp_env.update(envvars)
        popen = subprocess.Popen(
            [sys.executable, "-c", code],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=subp_env,
        )
        out, err = popen.communicate()
        if popen.returncode != 0:
            raise AssertionError(
                "process failed with code %s: \n"
                "stdout follows\n%s\n"
                "stderr follows\n%s\n"
                % (popen.returncode, out.decode(), err.decode()),
            )

    def check_hits(self, func, hits, misses=None):
        st = func.stats
        self.assertEqual(sum(st.cache_hits.values()), hits, st.cache_hits)
        if misses is not None:
            self.assertEqual(
                sum(st.cache_misses.values()), misses, st.cache_misses
            )


def check_access_is_preventable():
    # This exists to check whether it is possible to prevent access to
    # a file/directory through the use of `chmod 500`. If a user has
    # elevated rights (e.g. root) then writes are likely to be possible
    # anyway. Tests that require functioning access prevention are
    # therefore skipped based on the result of this check.
    tempdir = temp_directory("test_cache")
    test_dir = os.path.join(tempdir, "writable_test")
    os.mkdir(test_dir)
    # check a write is possible
    with open(os.path.join(test_dir, "write_ok"), "wt") as f:
        f.write("check1")
    # now forbid access
    os.chmod(test_dir, 0o500)
    try:
        with open(os.path.join(test_dir, "write_forbidden"), "wt") as f:
            f.write("check2")
        # access prevention is not possible
        return False
    except PermissionError:
        # Check that the cause of the exception is due to access/permission
        # as per
        # https://github.com/conda/conda/blob/4.5.0/conda/gateways/disk/permissions.py#L35-L37  # noqa: E501
        # errno reports access/perm fail so access prevention via
        # `chmod 500` works for this user.
        return True
    finally:
        os.chmod(test_dir, 0o775)
        shutil.rmtree(test_dir)


_access_preventable = check_access_is_preventable()
_access_msg = "Cannot create a directory to which writes are preventable"
skip_bad_access = unittest.skipUnless(_access_preventable, _access_msg)


@skip_on_cudasim("Simulator does not implement caching")
class CUDACachingTest(DispatcherCacheUsecasesTest):
    here = os.path.dirname(__file__)
    usecases_file = os.path.join(here, "cache_usecases.py")
    modname = "cuda_caching_test_fodder"

    def setUp(self):
        DispatcherCacheUsecasesTest.setUp(self)
        CUDATestCase.setUp(self)

    def tearDown(self):
        CUDATestCase.tearDown(self)
        DispatcherCacheUsecasesTest.tearDown(self)

    def test_caching(self):
        self.check_pycache(0)
        mod = self.import_module()
        self.check_pycache(0)

        f = mod.add_usecase
        self.assertPreciseEqual(f(2, 3), 6)
        self.check_pycache(2)  # 1 index, 1 data
        self.assertPreciseEqual(f(2.5, 3), 6.5)
        self.check_pycache(3)  # 1 index, 2 data
        self.check_hits(f.func, 0, 2)

        f = mod.record_return_aligned
        rec = f(mod.aligned_arr, 1)
        self.assertPreciseEqual(tuple(rec), (2, 43.5))

        f = mod.record_return_packed
        rec = f(mod.packed_arr, 1)
        self.assertPreciseEqual(tuple(rec), (2, 43.5))
        self.check_pycache(6)  # 2 index, 4 data
        self.check_hits(f.func, 0, 2)

        # Check the code runs ok from another process
        self.run_in_separate_process()

    def test_no_caching(self):
        mod = self.import_module()

        f = mod.add_nocache_usecase
        self.assertPreciseEqual(f(2, 3), 6)
        self.check_pycache(0)

    def test_many_locals(self):
        # Declaring many local arrays creates a very large LLVM IR, which
        # cannot be pickled due to the level of recursion it requires to
        # pickle. This test ensures that kernels with many locals (and
        # therefore large IR) can be cached. See Issue #8373:
        # https://github.com/numba/numba/issues/8373
        self.check_pycache(0)
        mod = self.import_module()
        f = mod.many_locals
        f[1, 1]()
        self.check_pycache(2)  # 1 index, 1 data

    def test_closure(self):
        mod = self.import_module()

        with warnings.catch_warnings():
            warnings.simplefilter("error", NumbaWarning)

            f = mod.closure1
            self.assertPreciseEqual(f(3), 6)  # 3 + 3 = 6
            f = mod.closure2
            self.assertPreciseEqual(f(3), 8)  # 3 + 5 = 8
            f = mod.closure3
            self.assertPreciseEqual(f(3), 10)  # 3 + 7 = 10
            f = mod.closure4
            self.assertPreciseEqual(f(3), 12)  # 3 + 9 = 12
            self.check_pycache(5)  # 1 nbi, 4 nbc

    def test_cache_reuse(self):
        mod = self.import_module()
        mod.add_usecase(2, 3)
        mod.add_usecase(2.5, 3.5)
        mod.outer_uncached(2, 3)
        mod.outer(2, 3)
        mod.record_return_packed(mod.packed_arr, 0)
        mod.record_return_aligned(mod.aligned_arr, 1)
        mod.simple_usecase_caller(2)
        mtimes = self.get_cache_mtimes()
        # Two signatures compiled
        self.check_hits(mod.add_usecase.func, 0, 2)

        mod2 = self.import_module()
        self.assertIsNot(mod, mod2)
        f = mod2.add_usecase
        f(2, 3)
        self.check_hits(f.func, 1, 0)
        f(2.5, 3.5)
        self.check_hits(f.func, 2, 0)

        # The files haven't changed
        self.assertEqual(self.get_cache_mtimes(), mtimes)

        self.run_in_separate_process()
        self.assertEqual(self.get_cache_mtimes(), mtimes)

    def test_cache_invalidate(self):
        mod = self.import_module()
        f = mod.add_usecase
        self.assertPreciseEqual(f(2, 3), 6)

        # This should change the functions' results
        with open(self.modfile, "a") as f:
            f.write("\nZ = 10\n")

        mod = self.import_module()
        f = mod.add_usecase
        self.assertPreciseEqual(f(2, 3), 15)

    def test_recompile(self):
        # Explicit call to recompile() should overwrite the cache
        mod = self.import_module()
        f = mod.add_usecase
        self.assertPreciseEqual(f(2, 3), 6)

        mod = self.import_module()
        f = mod.add_usecase
        mod.Z = 10
        self.assertPreciseEqual(f(2, 3), 6)
        f.func.recompile()
        self.assertPreciseEqual(f(2, 3), 15)

        # Freshly recompiled version is re-used from other imports
        mod = self.import_module()
        f = mod.add_usecase
        self.assertPreciseEqual(f(2, 3), 15)

    def test_same_names(self):
        # Function with the same names should still disambiguate
        mod = self.import_module()
        f = mod.renamed_function1
        self.assertPreciseEqual(f(2), 4)
        f = mod.renamed_function2
        self.assertPreciseEqual(f(2), 8)

    def _test_pycache_fallback(self):
        """
        With a disabled __pycache__, test there is a working fallback
        (e.g. on the user-wide cache dir)
        """
        mod = self.import_module()
        f = mod.add_usecase
        # Remove this function's cache files at the end, to avoid accumulation
        # across test calls.
        self.addCleanup(
            shutil.rmtree, f.func.stats.cache_path, ignore_errors=True
        )

        self.assertPreciseEqual(f(2, 3), 6)
        # It's a cache miss since the file was copied to a new temp location
        self.check_hits(f.func, 0, 1)

        # Test re-use
        mod2 = self.import_module()
        f = mod2.add_usecase
        self.assertPreciseEqual(f(2, 3), 6)
        self.check_hits(f.func, 1, 0)

        # The __pycache__ is empty (otherwise the test's preconditions
        # wouldn't be met)
        self.check_pycache(0)

    @skip_bad_access
    @unittest.skipIf(
        os.name == "nt", "cannot easily make a directory read-only on Windows"
    )
    def test_non_creatable_pycache(self):
        # Make it impossible to create the __pycache__ directory
        old_perms = os.stat(self.tempdir).st_mode
        os.chmod(self.tempdir, 0o500)
        self.addCleanup(os.chmod, self.tempdir, old_perms)

        self._test_pycache_fallback()

    @skip_bad_access
    @unittest.skipIf(
        os.name == "nt", "cannot easily make a directory read-only on Windows"
    )
    def test_non_writable_pycache(self):
        # Make it impossible to write to the __pycache__ directory
        pycache = os.path.join(self.tempdir, "__pycache__")
        os.mkdir(pycache)
        old_perms = os.stat(pycache).st_mode
        os.chmod(pycache, 0o500)
        self.addCleanup(os.chmod, pycache, old_perms)

        self._test_pycache_fallback()

    def test_cannot_cache_linking_libraries(self):
        link = str(test_data_dir / "jitlink.ptx")
        msg = "Cannot pickle CUDACodeLibrary with linking files"
        with self.assertRaisesRegex(RuntimeError, msg):

            @cuda.jit("void()", cache=True, link=[link])
            def f():
                pass

    def test_cannot_cache_captured_device_array(self):
        # Test that kernels capturing device arrays from closures cannot
        # be cached. The error can come from either NumbaPickler (for closure
        # variables) or CUDACodeLibrary._reduce_states (for referenced objects).
        host_data = np.array([1.0, 2.0, 3.0], dtype=np.float32)
        captured_arr = cuda.to_device(host_data)

        msg = "global device arrays"
        with self.assertRaisesRegex(PicklingError, msg):

            @cuda.jit(cache=True)
            def cached_kernel(output):
                i = cuda.grid(1)
                if i < output.size:
                    output[i] = captured_arr[i] * 2.0

            output = cuda.device_array(3, dtype=np.float32)
            cached_kernel[1, 3](output)

    def test_cannot_cache_global_device_array(self):
        # Test that kernels referencing module-level global device arrays
        # cannot be cached.
        global GLOBAL_DEVICE_ARRAY

        host_data = np.array([1.0, 2.0, 3.0], dtype=np.float32)
        GLOBAL_DEVICE_ARRAY = cuda.to_device(host_data)

        try:
            msg = "global device arrays"
            with self.assertRaisesRegex(PicklingError, msg):

                @cuda.jit(cache=True)
                def cached_kernel_global(output):
                    i = cuda.grid(1)
                    if i < output.size:
                        output[i] = GLOBAL_DEVICE_ARRAY[i] * 2.0

                output = cuda.device_array(3, dtype=np.float32)
                cached_kernel_global[1, 3](output)
        finally:
            GLOBAL_DEVICE_ARRAY = None


@skip_on_cudasim("Simulator does not implement caching")
class CUDACooperativeGroupTest(DispatcherCacheUsecasesTest):
    # See Issue #9432: https://github.com/numba/numba/issues/9432
    # If a cached function using CG sync was the first thing to compile,
    # the compile would fail.
    here = os.path.dirname(__file__)
    usecases_file = os.path.join(here, "cg_cache_usecases.py")
    modname = "cuda_cooperative_caching_test_fodder"

    def setUp(self):
        DispatcherCacheUsecasesTest.setUp(self)
        CUDATestCase.setUp(self)

    def tearDown(self):
        CUDATestCase.tearDown(self)
        DispatcherCacheUsecasesTest.tearDown(self)

    @skip_unless_cc_60
    @skip_if_cudadevrt_missing
    def test_cache_cg(self):
        # Functions using cooperative groups should be cacheable. See Issue
        # #8888: https://github.com/numba/numba/issues/8888
        self.check_pycache(0)
        mod = self.import_module()
        self.check_pycache(0)

        mod.cg_usecase(0)
        self.check_pycache(2)  # 1 index, 1 data

        # Check the code runs ok from another process
        self.run_in_separate_process()


@skip_on_cudasim("Simulator does not implement caching")
class CUDAAndCPUCachingTest(DispatcherCacheUsecasesTest):
    here = os.path.dirname(__file__)
    usecases_file = os.path.join(here, "cache_with_cpu_usecases.py")
    modname = "cuda_and_cpu_caching_test_fodder"

    def setUp(self):
        DispatcherCacheUsecasesTest.setUp(self)
        CUDATestCase.setUp(self)

    def tearDown(self):
        CUDATestCase.tearDown(self)
        DispatcherCacheUsecasesTest.tearDown(self)

    @skip_on_standalone_numba_cuda
    def test_cpu_and_cuda_targets(self):
        # The same function jitted for CPU and CUDA targets should maintain
        # separate caches for each target.
        self.check_pycache(0)
        mod = self.import_module()
        self.check_pycache(0)

        f_cpu = mod.assign_cpu
        f_cuda = mod.assign_cuda
        self.assertPreciseEqual(f_cpu(5), 5)
        self.check_pycache(2)  # 1 index, 1 data
        self.assertPreciseEqual(f_cuda(5), 5)
        self.check_pycache(3)  # 1 index, 2 data

        self.check_hits(f_cpu.func, 0, 1)
        self.check_hits(f_cuda.func, 0, 1)

        self.assertPreciseEqual(f_cpu(5.5), 5.5)
        self.check_pycache(4)  # 1 index, 3 data
        self.assertPreciseEqual(f_cuda(5.5), 5.5)
        self.check_pycache(5)  # 1 index, 4 data

        self.check_hits(f_cpu.func, 0, 2)
        self.check_hits(f_cuda.func, 0, 2)

    @skip_on_standalone_numba_cuda
    def test_cpu_and_cuda_reuse(self):
        # Existing cache files for the CPU and CUDA targets are reused.
        mod = self.import_module()
        mod.assign_cpu(5)
        mod.assign_cpu(5.5)
        mod.assign_cuda(5)
        mod.assign_cuda(5.5)

        mtimes = self.get_cache_mtimes()

        # Two signatures compiled
        self.check_hits(mod.assign_cpu.func, 0, 2)
        self.check_hits(mod.assign_cuda.func, 0, 2)

        mod2 = self.import_module()
        self.assertIsNot(mod, mod2)
        f_cpu = mod2.assign_cpu
        f_cuda = mod2.assign_cuda

        f_cpu(2)
        self.check_hits(f_cpu.func, 1, 0)
        f_cpu(2.5)
        self.check_hits(f_cpu.func, 2, 0)
        f_cuda(2)
        self.check_hits(f_cuda.func, 1, 0)
        f_cuda(2.5)
        self.check_hits(f_cuda.func, 2, 0)

        # The files haven't changed
        self.assertEqual(self.get_cache_mtimes(), mtimes)

        self.run_in_separate_process()
        self.assertEqual(self.get_cache_mtimes(), mtimes)


def get_different_cc_gpus():
    # Find two GPUs with different Compute Capabilities and return them as a
    # tuple. If two GPUs with distinct Compute Capabilities cannot be found,
    # then None is returned.
    first_gpu = cuda.gpus[0]
    with first_gpu:
        first_cc = cuda.current_context().device.compute_capability

    for gpu in cuda.gpus[1:]:
        with gpu:
            cc = cuda.current_context().device.compute_capability
            if cc != first_cc:
                return (first_gpu, gpu)

    return None


@skip_on_cudasim("Simulator does not implement caching")
class TestMultiCCCaching(DispatcherCacheUsecasesTest):
    here = os.path.dirname(__file__)
    usecases_file = os.path.join(here, "cache_usecases.py")
    modname = "cuda_multi_cc_caching_test_fodder"

    def setUp(self):
        DispatcherCacheUsecasesTest.setUp(self)
        CUDATestCase.setUp(self)

    def tearDown(self):
        CUDATestCase.tearDown(self)
        DispatcherCacheUsecasesTest.tearDown(self)

    def test_cache(self):
        gpus = get_different_cc_gpus()
        if not gpus:
            self.skipTest("Need two different CCs for multi-CC cache test")

        self.check_pycache(0)
        mod = self.import_module()
        self.check_pycache(0)

        # Step 1. Populate the cache with the first GPU
        with gpus[0]:
            f = mod.add_usecase
            self.assertPreciseEqual(f(2, 3), 6)
            self.check_pycache(2)  # 1 index, 1 data
            self.assertPreciseEqual(f(2.5, 3), 6.5)
            self.check_pycache(3)  # 1 index, 2 data
            self.check_hits(f.func, 0, 2)

            f = mod.record_return_aligned
            rec = f(mod.aligned_arr, 1)
            self.assertPreciseEqual(tuple(rec), (2, 43.5))

            f = mod.record_return_packed
            rec = f(mod.packed_arr, 1)
            self.assertPreciseEqual(tuple(rec), (2, 43.5))
            self.check_pycache(6)  # 2 index, 4 data
            self.check_hits(f.func, 0, 2)

        # Step 2. Run with the second GPU - under present behaviour this
        # doesn't further populate the cache.
        with gpus[1]:
            f = mod.add_usecase
            self.assertPreciseEqual(f(2, 3), 6)
            self.check_pycache(6)  # cache unchanged
            self.assertPreciseEqual(f(2.5, 3), 6.5)
            self.check_pycache(6)  # cache unchanged
            self.check_hits(f.func, 0, 2)

            f = mod.record_return_aligned
            rec = f(mod.aligned_arr, 1)
            self.assertPreciseEqual(tuple(rec), (2, 43.5))

            f = mod.record_return_packed
            rec = f(mod.packed_arr, 1)
            self.assertPreciseEqual(tuple(rec), (2, 43.5))
            self.check_pycache(6)  # cache unchanged
            self.check_hits(f.func, 0, 2)

        # Step 3. Run in a separate module with the second GPU - this populates
        # the cache for the second CC.
        mod2 = self.import_module()
        self.assertIsNot(mod, mod2)

        with gpus[1]:
            f = mod2.add_usecase
            self.assertPreciseEqual(f(2, 3), 6)
            self.check_pycache(7)  # 2 index, 5 data
            self.assertPreciseEqual(f(2.5, 3), 6.5)
            self.check_pycache(8)  # 2 index, 6 data
            self.check_hits(f.func, 0, 2)

            f = mod2.record_return_aligned
            rec = f(mod.aligned_arr, 1)
            self.assertPreciseEqual(tuple(rec), (2, 43.5))

            f = mod2.record_return_packed
            rec = f(mod.packed_arr, 1)
            self.assertPreciseEqual(tuple(rec), (2, 43.5))
            self.check_pycache(10)  # 2 index, 8 data
            self.check_hits(f.func, 0, 2)

        # The following steps check that we can use the NVVM IR loaded from the
        # cache to generate PTX for a different compute capability to the
        # cached cubin's CC. To check this, we create another module that loads
        # the cached version containing a cubin for GPU 1. There will be no
        # cubin for GPU 0, so when we try to use it the PTX must be generated.

        mod3 = self.import_module()
        self.assertIsNot(mod, mod3)

        # Step 4. Run with GPU 1 and get a cache hit, loading the cache created
        # during Step 3.
        with gpus[1]:
            f = mod3.add_usecase
            self.assertPreciseEqual(f(2, 3), 6)
            self.assertPreciseEqual(f(2.5, 3), 6.5)

            f = mod3.record_return_aligned
            rec = f(mod.aligned_arr, 1)
            self.assertPreciseEqual(tuple(rec), (2, 43.5))

            f = mod3.record_return_packed
            rec = f(mod.packed_arr, 1)
            self.assertPreciseEqual(tuple(rec), (2, 43.5))

        # Step 5. Run with GPU 0 using the module from Step 4, to force PTX
        # generation from cached NVVM IR.
        with gpus[0]:
            f = mod3.add_usecase
            self.assertPreciseEqual(f(2, 3), 6)
            self.assertPreciseEqual(f(2.5, 3), 6.5)

            f = mod3.record_return_aligned
            rec = f(mod.aligned_arr, 1)
            self.assertPreciseEqual(tuple(rec), (2, 43.5))

            f = mod3.record_return_packed
            rec = f(mod.packed_arr, 1)
            self.assertPreciseEqual(tuple(rec), (2, 43.5))


def child_initializer():
    # Disable occupancy and implicit copy warnings in processes in a
    # multiprocessing pool.
    from numba.cuda.core import config

    config.CUDA_LOW_OCCUPANCY_WARNINGS = 0
    config.CUDA_WARN_ON_IMPLICIT_COPY = 0


@skip_on_cudasim("Simulator does not implement caching")
class TestMultiprocessCache(DispatcherCacheUsecasesTest):
    here = os.path.dirname(__file__)
    usecases_file = os.path.join(here, "cache_usecases.py")
    modname = "cuda_mp_caching_test_fodder"

    def setUp(self):
        DispatcherCacheUsecasesTest.setUp(self)
        CUDATestCase.setUp(self)

    def tearDown(self):
        CUDATestCase.tearDown(self)
        DispatcherCacheUsecasesTest.tearDown(self)

    def test_multiprocessing(self):
        # Check caching works from multiple processes at once (#2028)
        mod = self.import_module()
        # Calling a pure Python caller of the JIT-compiled function is
        # necessary to reproduce the issue.
        f = mod.simple_usecase_caller
        n = 3
        try:
            ctx = multiprocessing.get_context("spawn")
        except AttributeError:
            ctx = multiprocessing

        pool = ctx.Pool(n, child_initializer)

        try:
            res = sum(pool.imap(f, range(n)))
        finally:
            pool.close()
        self.assertEqual(res, n * (n - 1) // 2)


@skip_on_cudasim("Simulator does not implement the CUDACodeLibrary")
class TestCUDACodeLibrary(CUDATestCase):
    # For tests of miscellaneous CUDACodeLibrary behaviour that we wish to
    # explicitly check

    def test_cannot_serialize_unfinalized(self):
        # The CUDA codegen failes to import under the simulator, so we cannot
        # import it at the top level
        from numba.cuda.codegen import CUDACodeLibrary

        # Usually a CodeLibrary requires a real CodeGen, but since we don't
        # interact with it, anything will do
        codegen = object()
        name = "library"
        cl = CUDACodeLibrary(codegen, name)
        with self.assertRaisesRegex(RuntimeError, "Cannot pickle unfinalized"):
            cl._reduce_states()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from ctypes import c_int, sizeof

from numba.cuda.cudadrv.driver import host_to_device, device_to_host, driver
from cuda.core.experimental import (
    LaunchConfig,
    Stream as ExperimentalStream,
    launch,
)

from numba import cuda
from numba.cuda.cudadrv import devices
from numba.cuda.testing import unittest, CUDATestCase
from numba.cuda.testing import skip_on_cudasim
import contextlib

from cuda.core.experimental import Device

ptx1 = """
    .version 1.4
    .target sm_10, map_f64_to_f32

    .entry _Z10helloworldPi (
    .param .u64 __cudaparm__Z10helloworldPi_A)
    {
    .reg .u32 %r<3>;
    .reg .u64 %rd<6>;
    .loc	14	4	0
$LDWbegin__Z10helloworldPi:
    .loc	14	6	0
    cvt.s32.u16 	%r1, %tid.x;
    ld.param.u64 	%rd1, [__cudaparm__Z10helloworldPi_A];
    cvt.u64.u16 	%rd2, %tid.x;
    mul.lo.u64 	%rd3, %rd2, 4;
    add.u64 	%rd4, %rd1, %rd3;
    st.global.s32 	[%rd4+0], %r1;
    .loc	14	7	0
    exit;
$LDWend__Z10helloworldPi:
    } // _Z10helloworldPi
"""

ptx2 = """
.version 3.0
.target sm_20
.address_size 64

    .file	1 "/tmp/tmpxft_000012c7_00000000-9_testcuda.cpp3.i"
    .file	2 "testcuda.cu"

.entry _Z10helloworldPi(
    .param .u64 _Z10helloworldPi_param_0
)
{
    .reg .s32 	%r<3>;
    .reg .s64 	%rl<5>;


    ld.param.u64 	%rl1, [_Z10helloworldPi_param_0];
    cvta.to.global.u64 	%rl2, %rl1;
    .loc 2 6 1
    mov.u32 	%r1, %tid.x;
    mul.wide.u32 	%rl3, %r1, 4;
    add.s64 	%rl4, %rl2, %rl3;
    st.global.u32 	[%rl4], %r1;
    .loc 2 7 2
    ret;
}
"""


@skip_on_cudasim("CUDA Driver API unsupported in the simulator")
class TestCudaDriver(CUDATestCase):
    def setUp(self):
        super().setUp()
        self.assertTrue(len(devices.gpus) > 0)
        self.context = devices.get_context()
        device = self.context.device
        ccmajor, _ = device.compute_capability
        if ccmajor >= 2:
            self.ptx = ptx2
        else:
            self.ptx = ptx1

    def tearDown(self):
        super().tearDown()
        del self.context

    def test_cuda_driver_basic(self):
        module = self.context.create_module_ptx(self.ptx)
        function = module.get_function("_Z10helloworldPi")

        array = (c_int * 100)()

        memory = self.context.memalloc(sizeof(array))
        host_to_device(memory, array, sizeof(array))

        ptr = memory.device_ctypes_pointer

        config = LaunchConfig(
            grid=(1, 1, 1),
            block=(100, 1, 1),
            shmem_size=0,
            cooperative_launch=False,
        )
        exp_stream = ExperimentalStream.from_handle(0)
        launch(exp_stream, config, function.kernel, ptr)

        device_to_host(array, memory, sizeof(array))
        for i, v in enumerate(array):
            self.assertEqual(i, v)

        module.unload()

    def test_cuda_driver_stream_operations(self):
        from numba.cuda.cudadrv.driver import _to_core_stream

        module = self.context.create_module_ptx(self.ptx)
        function = module.get_function("_Z10helloworldPi")

        array = (c_int * 100)()

        stream = self.context.create_stream()

        with stream.auto_synchronize():
            memory = self.context.memalloc(sizeof(array))
            host_to_device(memory, array, sizeof(array), stream=stream)

            ptr = memory.device_ctypes_pointer

            config = LaunchConfig(
                grid=(1, 1, 1),
                block=(100, 1, 1),
                shmem_size=0,
                cooperative_launch=False,
            )
            # Convert numba Stream to ExperimentalStream
            launch(_to_core_stream(stream), config, function.kernel, ptr)

        device_to_host(array, memory, sizeof(array), stream=stream)

        for i, v in enumerate(array):
            self.assertEqual(i, v)

    def test_cuda_core_stream_operations(self):
        module = self.context.create_module_ptx(self.ptx)
        function = module.get_function("_Z10helloworldPi")
        array = (c_int * 100)()
        dev = Device()
        dev.set_current()
        stream = dev.create_stream()

        @contextlib.contextmanager
        def auto_synchronize(stream):
            try:
                yield stream
            finally:
                stream.sync()

        with auto_synchronize(stream):
            memory = self.context.memalloc(sizeof(array))
            host_to_device(memory, array, sizeof(array), stream=stream)

            ptr = memory.device_ctypes_pointer

            config = LaunchConfig(
                grid=(1, 1, 1),
                block=(100, 1, 1),
                shmem_size=0,
                cooperative_launch=False,
            )
            launch(stream, config, function.kernel, ptr)

            device_to_host(array, memory, sizeof(array), stream=stream)
        for i, v in enumerate(array):
            self.assertEqual(i, v)

    def test_cuda_core_stream_launch_user_facing(self):
        @cuda.jit
        def kernel(a):
            idx = cuda.grid(1)
            if idx < len(a):
                a[idx] = idx

        dev = Device()
        dev.set_current()
        stream = dev.create_stream()

        ary = cuda.to_device([0] * 100, stream=stream)
        stream.sync()

        kernel[1, 100, stream](ary)
        stream.sync()

        result = ary.copy_to_host(stream=stream)
        for i, v in enumerate(result):
            self.assertEqual(i, v)

    def test_cuda_driver_default_stream(self):
        # Test properties of the default stream
        ds = self.context.get_default_stream()
        self.assertIn("Default CUDA stream", repr(ds))
        self.assertEqual(0, int(ds))
        # bool(stream) is the check that is done in memcpy to decide if async
        # version should be used. So the default (0) stream should be true-ish
        # even though 0 is usually false-ish in Python.
        self.assertTrue(ds)
        self.assertFalse(ds.external)

    def test_cuda_driver_legacy_default_stream(self):
        # Test properties of the legacy default stream
        ds = self.context.get_legacy_default_stream()
        self.assertIn("Legacy default CUDA stream", repr(ds))
        self.assertEqual(1, int(ds))
        self.assertTrue(ds)
        self.assertFalse(ds.external)

    def test_cuda_driver_per_thread_default_stream(self):
        # Test properties of the per-thread default stream
        ds = self.context.get_per_thread_default_stream()
        self.assertIn("Per-thread default CUDA stream", repr(ds))
        self.assertEqual(2, int(ds))
        self.assertTrue(ds)
        self.assertFalse(ds.external)

    def test_cuda_driver_stream(self):
        # Test properties of non-default streams
        s = self.context.create_stream()
        self.assertIn("CUDA stream", repr(s))
        self.assertNotIn("Default", repr(s))
        self.assertNotIn("External", repr(s))
        self.assertNotEqual(0, int(s))
        self.assertTrue(s)
        self.assertFalse(s.external)

    def test_cuda_driver_external_stream(self):
        # Test properties of a stream created from an external stream object.
        # We use the driver API directly to create a stream, to emulate an
        # external library creating a stream
        handle = driver.cuStreamCreate(0)
        ptr = int(handle)
        s = self.context.create_external_stream(ptr)

        self.assertIn("External CUDA stream", repr(s))
        # Ensure neither "Default" nor "default"
        self.assertNotIn("efault", repr(s))
        self.assertEqual(ptr, int(s))
        self.assertTrue(s)
        self.assertTrue(s.external)

    def test_cuda_driver_occupancy(self):
        module = self.context.create_module_ptx(self.ptx)
        function = module.get_function("_Z10helloworldPi")

        value = self.context.get_active_blocks_per_multiprocessor(
            function, 128, 128
        )
        self.assertTrue(value > 0)

        def b2d(bs):
            return bs

        grid, block = self.context.get_max_potential_block_size(
            function, b2d, 128, 128
        )
        self.assertTrue(grid > 0)
        self.assertTrue(block > 0)

    def test_cuda_cache_config(self):
        from numba import types
        import numpy as np

        sig = (types.float32[::1], types.float32[::1])

        @cuda.jit(sig)
        def add_one(r, x):
            i = cuda.grid(1)
            if i < len(r):
                r[i] = x[i] + 1

        kernel = add_one.overloads[sig]
        cufunc = kernel._codelibrary.get_cufunc()

        configs_to_test = [
            ("prefer_shared", dict(prefer_shared=True)),
            ("prefer_cache", dict(prefer_cache=True)),
            ("prefer_equal", dict(prefer_equal=True)),
            ("default", dict()),
        ]

        for name, kwargs in configs_to_test:
            with self.subTest(config=name):
                try:
                    cufunc.cache_config(**kwargs)
                except Exception as e:
                    self.fail(f"cache_config({name}) failed: {e}")

        x = np.array([1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32)
        r = np.zeros_like(x)

        d_x = cuda.to_device(x)
        d_r = cuda.to_device(r)

        cufunc.cache_config(prefer_shared=True)
        add_one[1, 5](d_r, d_x)

        result = d_r.copy_to_host()
        expected = x + 1

        np.testing.assert_array_almost_equal(
            result,
            expected,
            err_msg="Kernel produced incorrect results after cache_config",
        )

    def test_cuda_set_shared_memory_carveout(self):
        from numba import types
        import numpy as np

        sig = (types.float32[::1], types.float32[::1])

        @cuda.jit(sig)
        def add_one(r, x):
            i = cuda.grid(1)
            if i < len(r):
                r[i] = x[i] + 1

        kernel = add_one.overloads[sig]
        cufunc = kernel._codelibrary.get_cufunc()

        # valid carveout values
        carveout_values = [-1, 0, 50, 100]
        for value in carveout_values:
            with self.subTest(carveout=value):
                try:
                    cufunc.set_shared_memory_carveout(value)
                except Exception as e:
                    self.fail(
                        f"set_shared_memory_carveout({value}) failed: {e}"
                    )

        # invalid carveout values
        invalid_values = [-2, 101, 150]
        for value in invalid_values:
            with self.subTest(invalid_carveout=value):
                with self.assertRaises(ValueError):
                    cufunc.set_shared_memory_carveout(value)

        # test the kernel
        x = np.array([1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32)
        r = np.zeros_like(x)

        d_x = cuda.to_device(x)
        d_r = cuda.to_device(r)

        cufunc.set_shared_memory_carveout(75)
        add_one[1, 5](d_r, d_x)

        result = d_r.copy_to_host()
        expected = x + 1

        np.testing.assert_array_almost_equal(
            result,
            expected,
            err_msg="Kernel produced incorrect results after set_shared_memory_carveout",
        )


class TestDevice(CUDATestCase):
    def test_device_get_uuid(self):
        # A device UUID looks like:
        #
        #     GPU-e6489c45-5b68-3b03-bab7-0e7c8e809643
        #
        # To test, we construct an RE that matches this form and verify that
        # the returned UUID matches.
        #
        # Device UUIDs may not conform to parts of the UUID specification (RFC
        # 4122) pertaining to versions and variants, so we do not extract and
        # validate the values of these bits.

        h = "[0-9a-f]{%d}"
        h4 = h % 4
        h8 = h % 8
        h12 = h % 12
        uuid_format = f"^GPU-{h8}-{h4}-{h4}-{h4}-{h12}$"

        dev = devices.get_context().device
        self.assertRegex(dev.uuid, uuid_format)


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import pathlib
import sys

from setuptools import setup, Extension
from setuptools.command.build_py import build_py
from setuptools.command.editable_wheel import editable_wheel, _TopLevelFinder
from setuptools.command.build_ext import build_ext

REDIRECTOR_PTH = "_numba_cuda_redirector.pth"
REDIRECTOR_PY = "_numba_cuda_redirector.py"
SITE_PACKAGES = pathlib.Path("site-packages")


def get_version():
    """Read version from VERSION file."""
    version_file = pathlib.Path(__file__).parent / "numba_cuda" / "VERSION"
    return version_file.read_text().strip()


def get_ext_modules():
    """
    Return a list of Extension instances for the setup() call.
    """
    # Note we don't import NumPy at the toplevel, since setup.py
    # should be able to run without NumPy for pip to discover the
    # build dependencies. Need NumPy headers and libm linkage.
    import numpy as np

    np_compile_args = {
        "include_dirs": [
            np.get_include(),
        ],
    }
    if sys.platform != "win32":
        np_compile_args["libraries"] = [
            "m",
        ]

    install_name_tool_fixer = []
    if sys.platform == "darwin":
        install_name_tool_fixer = ["-headerpad_max_install_names"]

    ext_mviewbuf = Extension(
        name="numba_cuda.numba.cuda.cext.mviewbuf",
        extra_link_args=install_name_tool_fixer,
        sources=["numba_cuda/numba/cuda/cext/mviewbuf.c"],
    )

    dispatcher_sources = [
        "numba_cuda/numba/cuda/cext/_dispatcher.cpp",
        "numba_cuda/numba/cuda/cext/_typeof.cpp",
        "numba_cuda/numba/cuda/cext/_hashtable.cpp",
        "numba_cuda/numba/cuda/cext/typeconv.cpp",
    ]
    ext_dispatcher = Extension(
        name="numba_cuda.numba.cuda.cext._dispatcher",
        sources=dispatcher_sources,
        depends=[
            "numba_cuda/numba/cuda/cext/_pymodule.h",
            "numba_cuda/numba/cuda/cext/_typeof.h",
            "numba_cuda/numba/cuda/cext/_hashtable.h",
        ],
        extra_compile_args=["-std=c++11"],
        **np_compile_args,
    )

    ext_typeconv = Extension(
        name="numba_cuda.numba.cuda.cext._typeconv",
        sources=[
            "numba_cuda/numba/cuda/cext/typeconv.cpp",
            "numba_cuda/numba/cuda/cext/_typeconv.cpp",
        ],
        depends=["numba_cuda/numba/cuda/cext/_pymodule.h"],
        extra_compile_args=["-std=c++11"],
    )

    # Append our cext dir to include_dirs
    ext_dispatcher.include_dirs.append("numba_cuda/numba/cuda/cext")

    ext_helperlib = Extension(
        name="numba_cuda.numba.cuda.cext._helperlib",
        sources=["numba_cuda/numba/cuda/cext/_helpermod.c"],
        depends=[
            "numba_cuda/numba/cuda/cext/_pymodule.h",
            "numba_cuda/numba/cuda/cext/_helperlib.c",
        ],
        include_dirs=["numba_cuda/numba/cuda/cext"],
    )

    return [
        ext_dispatcher,
        ext_typeconv,
        ext_helperlib,
        ext_mviewbuf,
    ]


def is_building():
    """
    Parse the setup.py command and return whether a build is requested.
    If False is returned, only an informational command is run.
    If True is returned, information about C extensions will have to
    be passed to the setup() function.
    """
    if len(sys.argv) < 2:
        # User forgot to give an argument probably, let setuptools handle that.
        return True

    build_commands = [
        "build",
        "build_py",
        "build_ext",
        "build_clib",
        "build_scripts",
        "install",
        "install_lib",
        "install_headers",
        "install_scripts",
        "install_data",
        "sdist",
        "bdist",
        "bdist_dumb",
        "bdist_rpm",
        "bdist_wininst",
        "check",
        "build_docs",
        "bdist_wheel",
        "bdist_egg",
        "develop",
        "easy_install",
        "test",
        "editable_wheel",
    ]
    return any(bc in sys.argv[1:] for bc in build_commands)


# Adapted from https://stackoverflow.com/a/71137790
class build_py_with_redirector(build_py):  # noqa: N801
    """Include the redirector files in the generated wheel."""

    def copy_redirector_file(self, source, destination="."):
        destination = pathlib.Path(self.build_lib) / destination
        self.copy_file(str(source), str(destination), preserve_mode=0)

    def run(self):
        super().run()
        self.copy_redirector_file(SITE_PACKAGES / REDIRECTOR_PTH)
        self.copy_redirector_file(SITE_PACKAGES / REDIRECTOR_PY)

    def get_source_files(self):
        src = super().get_source_files()
        src.extend(
            [
                str(SITE_PACKAGES / REDIRECTOR_PTH),
                str(SITE_PACKAGES / REDIRECTOR_PY),
            ]
        )
        return src

    def get_output_mapping(self):
        mapping = super().get_output_mapping()
        build_lib = pathlib.Path(self.build_lib)
        mapping[str(build_lib / REDIRECTOR_PTH)] = REDIRECTOR_PTH
        mapping[str(build_lib / REDIRECTOR_PY)] = REDIRECTOR_PY
        return mapping


class TopLevelFinderWithRedirector(_TopLevelFinder):
    """Include the redirector files in the editable wheel."""

    def get_implementation(self):
        for item in super().get_implementation():
            yield item

        with open(SITE_PACKAGES / REDIRECTOR_PTH) as f:
            yield (REDIRECTOR_PTH, f.read())

        with open(SITE_PACKAGES / REDIRECTOR_PY) as f:
            yield (REDIRECTOR_PY, f.read())


class editable_wheel_with_redirector(editable_wheel):
    def _select_strategy(self, name, tag, build_lib):
        # The default mode is "lenient" - others are "strict" and "compat".
        # "compat" is deprecated. "strict" creates a tree of links to files in
        # the repo. It could be implemented, but we only handle the default
        # case for now.
        if self.mode is not None and self.mode != "lenient":
            raise RuntimeError(
                "Only lenient mode is supported for editable "
                f"install. Current mode is {self.mode}"
            )

        return TopLevelFinderWithRedirector(self.distribution, name)


cmdclass = {}

numba_be_user_options = [
    ("werror", None, "Build extensions with -Werror"),
    ("wall", None, "Build extensions with -Wall"),
    ("noopt", None, "Build extensions without optimization"),
]


class NumbaBuildExt(build_ext):
    user_options = build_ext.user_options + numba_be_user_options
    boolean_options = build_ext.boolean_options + ["werror", "wall", "noopt"]

    def initialize_options(self):
        super().initialize_options()
        self.werror = 0
        self.wall = 0
        self.noopt = 0

    def run(self):
        extra_compile_args = []
        if self.noopt:
            if sys.platform == "win32":
                extra_compile_args.append("/Od")
            else:
                extra_compile_args.append("-O0")
        if self.werror:
            extra_compile_args.append("-Werror")
        if self.wall:
            extra_compile_args.append("-Wall")
        for ext in self.extensions:
            ext.extra_compile_args.extend(extra_compile_args)

        super().run()


cmdclass["build_ext"] = NumbaBuildExt
cmdclass["build_py"] = build_py_with_redirector
cmdclass["editable_wheel"] = editable_wheel_with_redirector

if is_building():
    ext_modules = get_ext_modules()
else:
    ext_modules = []

setup(
    cmdclass=cmdclass,
    ext_modules=ext_modules,
)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import numpy as np
from numba.cuda import vectorize
from numba.cuda.testing import skip_on_cudasim, CUDATestCase
import unittest


@skip_on_cudasim("ufunc API unsupported in the simulator")
class TestVectorizeComplex(CUDATestCase):
    def test_vectorize_complex(self):
        @vectorize(["complex128(complex128)"], target="cuda")
        def vcomp(a):
            return a * a + 1.0

        A = np.arange(5, dtype=np.complex128)
        B = vcomp(A)
        self.assertTrue(np.allclose(A * A + 1.0, B))


if __name__ == "__main__":
    unittest.main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Tests for capturing device arrays (objects implementing __cuda_array_interface__)
from global scope in CUDA kernels and device functions.

This tests the capture of arrays that implement __cuda_array_interface__:
- Numba device arrays (cuda.to_device)
- ForeignArray (wrapper implementing __cuda_array_interface__)
"""

import numpy as np

from numba import cuda
from numba.cuda.testing import unittest, CUDATestCase, ForeignArray
from numba.cuda.testing import skip_on_cudasim


def make_numba_array(host_arr):
    """Create a Numba device array from host array."""
    return cuda.to_device(host_arr)


def make_foreign_array(host_arr):
    """Create a ForeignArray wrapping a Numba device array."""
    return ForeignArray(cuda.to_device(host_arr))


def get_host_data(arr):
    """Copy array data back to host."""
    if isinstance(arr, ForeignArray):
        return arr._arr.copy_to_host()
    return arr.copy_to_host()


# Array factories to test: (name, factory)
ARRAY_FACTORIES = [
    ("numba_device", make_numba_array),
    ("foreign", make_foreign_array),
]


@skip_on_cudasim("Global device array capture not supported in simulator")
class TestDeviceArrayCapture(CUDATestCase):
    """Test capturing device arrays from global scope."""

    def test_basic_capture(self):
        """Test basic global capture with different array types."""
        for name, make_array in ARRAY_FACTORIES:
            with self.subTest(array_type=name):
                host_data = np.array(
                    [1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32
                )
                global_array = make_array(host_data)

                @cuda.jit(device=True)
                def read_global(idx):
                    return global_array[idx]

                @cuda.jit
                def kernel(output):
                    i = cuda.grid(1)
                    if i < output.size:
                        output[i] = read_global(i)

                n = len(host_data)
                output = cuda.device_array(n, dtype=np.float32)
                kernel[1, n](output)

                result = output.copy_to_host()
                np.testing.assert_array_equal(result, host_data)

    def test_computation(self):
        """Test captured global arrays used in computations."""
        for name, make_array in ARRAY_FACTORIES:
            with self.subTest(array_type=name):
                host_data = np.array(
                    [1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32
                )
                global_array = make_array(host_data)

                @cuda.jit(device=True)
                def double_global_value(idx):
                    return global_array[idx] * 2.0

                @cuda.jit
                def kernel(output):
                    i = cuda.grid(1)
                    if i < output.size:
                        output[i] = double_global_value(i)

                n = len(host_data)
                output = cuda.device_array(n, dtype=np.float32)
                kernel[1, n](output)

                result = output.copy_to_host()
                expected = host_data * 2.0
                np.testing.assert_array_equal(result, expected)

    def test_mutability(self):
        """Test that captured arrays can be written to (mutability)."""
        for name, make_array in ARRAY_FACTORIES:
            with self.subTest(array_type=name):
                host_data = np.zeros(5, dtype=np.float32)
                mutable_array = make_array(host_data)

                @cuda.jit
                def write_kernel():
                    i = cuda.grid(1)
                    if i < 5:
                        mutable_array[i] = float(i + 1)

                write_kernel[1, 5]()

                result = get_host_data(mutable_array)
                expected = np.array([1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32)
                np.testing.assert_array_equal(result, expected)

    def test_multiple_arrays(self):
        """Test capturing multiple arrays from globals."""
        for name, make_array in ARRAY_FACTORIES:
            with self.subTest(array_type=name):
                host_a = np.array([1.0, 2.0, 3.0], dtype=np.float32)
                host_b = np.array([10.0, 20.0, 30.0], dtype=np.float32)
                arr_a = make_array(host_a)
                arr_b = make_array(host_b)

                @cuda.jit(device=True)
                def add_globals(idx):
                    return arr_a[idx] + arr_b[idx]

                @cuda.jit
                def kernel(output):
                    i = cuda.grid(1)
                    if i < output.size:
                        output[i] = add_globals(i)

                output = cuda.device_array(3, dtype=np.float32)
                kernel[1, 3](output)

                result = output.copy_to_host()
                expected = np.array([11.0, 22.0, 33.0], dtype=np.float32)
                np.testing.assert_array_equal(result, expected)

    def test_multidimensional(self):
        """Test capturing multidimensional arrays."""
        for name, make_array in ARRAY_FACTORIES:
            with self.subTest(array_type=name):
                host_2d = np.array(
                    [[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], dtype=np.float32
                )
                arr_2d = make_array(host_2d)

                @cuda.jit(device=True)
                def read_2d(row, col):
                    return arr_2d[row, col]

                @cuda.jit
                def kernel(output):
                    i = cuda.grid(1)
                    if i < 6:
                        row = i // 2
                        col = i % 2
                        output[i] = read_2d(row, col)

                output = cuda.device_array(6, dtype=np.float32)
                kernel[1, 6](output)

                result = output.copy_to_host()
                expected = host_2d.flatten()
                np.testing.assert_array_equal(result, expected)

    def test_dtypes(self):
        """Test capturing arrays with different dtypes."""
        dtypes = [
            (np.int32, [10, 20, 30, 40]),
            (np.float64, [1.5, 2.5, 3.5, 4.5]),
        ]

        for name, make_array in ARRAY_FACTORIES:
            for dtype, values in dtypes:
                with self.subTest(array_type=name, dtype=dtype):
                    host_data = np.array(values, dtype=dtype)
                    global_arr = make_array(host_data)

                    @cuda.jit(device=True)
                    def read_arr(idx):
                        return global_arr[idx]

                    @cuda.jit
                    def kernel(output):
                        i = cuda.grid(1)
                        if i < output.size:
                            output[i] = read_arr(i)

                    output = cuda.device_array(len(host_data), dtype=dtype)
                    kernel[1, len(host_data)](output)
                    np.testing.assert_array_equal(
                        output.copy_to_host(), host_data
                    )

    def test_direct_kernel_access(self):
        """Test direct kernel access (not via device function)."""
        for name, make_array in ARRAY_FACTORIES:
            with self.subTest(array_type=name):
                host_data = np.array([7.0, 8.0, 9.0], dtype=np.float32)
                global_direct = make_array(host_data)

                @cuda.jit
                def direct_access_kernel(output):
                    i = cuda.grid(1)
                    if i < output.size:
                        output[i] = global_direct[i] + 1.0

                output = cuda.device_array(3, dtype=np.float32)
                direct_access_kernel[1, 3](output)

                result = output.copy_to_host()
                expected = np.array([8.0, 9.0, 10.0], dtype=np.float32)
                np.testing.assert_array_equal(result, expected)

    def test_zero_dimensional(self):
        """Test capturing 0-D (scalar) device arrays."""
        for name, make_array in ARRAY_FACTORIES:
            with self.subTest(array_type=name):
                host_0d = np.array(42.0, dtype=np.float32)
                global_0d = make_array(host_0d)

                @cuda.jit
                def kernel_0d(output):
                    output[()] = global_0d[()] * 2.0

                output = cuda.device_array((), dtype=np.float32)
                kernel_0d[1, 1](output)

                result = output.copy_to_host()
                expected = 84.0
                self.assertEqual(result, expected)


if __name__ == "__main__":
    unittest.main()
